/*
 * ECE 18-447, Spring 2014
 *
 * ARM pipeline timing simulator
 *
 * Varun Kohli, 2014
 */

#include "pipe.h"
#include "shell.h"
#include "arm.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>


//#define DEBUG                                                                  

/* Pipeline helper functions */

/* debug */
void print_op(Pipe_Op *op)
{
  if (op)
    printf("OP (PC=%08x inst=%08x) src1=R%d (%08x) src2=R%d (%08x) dst=R%d v\
alid %d (%08x) is_mem=%d  CPSR=%08x\n",
	   op->pc, op->instruction, op->reg_src1, op->reg_src1_value, op->reg_src2, 
	   op->reg_src2_value, op->reg_dst, op->data_ready,
	   op->reg_dst_value, op->is_mem, op->CPSR);
  else
    printf("(null)\n");
}

int32_t sign_extend_24to30(int32_t c)
{
  return (c & 0x800000) ? (c | 0x3f800000) : c;
}

uint8_t mem_read_8(uint32_t addr)
{
  uint32_t word = mem_read_32(addr & ~3);
  switch (addr & 3) {
    case 0: return (word >>  0) & 0xFF;
    case 1: return (word >>  8) & 0xFF;
    case 2: return (word >> 16) & 0xFF;
    case 3: return (word >> 24) & 0xFF;
    default: return 0;
  }
}

void mem_write_8(uint32_t addr, uint8_t value)
{
  uint32_t word = mem_read_32(addr & ~3);
  
  switch (addr & 3) {
    case 0: word = (word & 0xFFFFFF00) | ((value & 0xFF) <<  0); break;
    case 1: word = (word & 0xFFFF00FF) | ((value & 0xFF) <<  8); break;
    case 2: word = (word & 0xFF00FFFF) | ((value & 0xFF) << 16); break;
    case 3: word = (word & 0x00FFFFFF) | ((value & 0xFF) << 24); break;
    default: break;
  }
  
  mem_write_32(addr & ~3, word);
}

uint8_t cond_is_met(Pipe_Op* op)
{
  assert(op != NULL);
  switch(op->cond) {
    case EQ:
      if(get_Z()) return 1;
      else return 0;
    case NE:
      if(!get_Z()) return 1;
      else return 0;
    case CS:
      if(get_C()) return 1;
      else return 0;
    case CC:
      if(!get_C()) return 1;
      else return 0;
    case MI:
      if(get_N()) return 1;
      else return 0;
    case PL:
      if(!get_N()) return 1;
      else return 0;
    case VS:
      if(get_V()) return 1;
      else return 0;
    case VC:
      if(!get_V()) return 1;
      else return 0;
    case HI:
      if(get_C() && !get_Z()) return 1;
      else return 0;
    case LS:
      if(!get_C() || get_Z()) return 1;
      else return 0;
    case GE:
      if((get_N()) == (get_V())) return 1;
      else return 0;
    case LT:
      if((get_N()) != (get_V())) return 1;
      else return 0;
    case GT:
      if(((get_N()) == (get_V())) && (!get_Z())) return 1;
      else return 0;
    case LE:
      if(((get_N()) != (get_V())) || (get_Z())) return 1;
      else return 0;
    }
  return 1;
}

/* global pipeline state */
Pipe_State pipe;

void pipe_init() 
{
  memset(&pipe, 0, sizeof(Pipe_State));
  pipe.PC = 0x00400000;
}

/* Each execution does one cycle of the pipeline */
void cycle_pipe()
{
#ifdef DEBUG
    printf("\n\n----\n\nPIPELINE:\n");
    printf("DCODE: "); print_op(pipe.decode_op);
    printf("EXEC : "); print_op(pipe.execute_op);
    printf("MEM  : "); print_op(pipe.mem_op);
    printf("WB   : "); print_op(pipe.wb_op);
    printf("Cycles: %d\n", stat_cycles);
    printf("\n");
#endif

  pipe_stage_wb();
  pipe_stage_mem();
  pipe_stage_ex();
  pipe_stage_decode();
  pipe_stage_fetch();
  
  if(pipe.branch_recover) {
#ifdef DEBUG
    printf("branch recovery: new dest %08x flush %d stages \n", pipe.branch_dest, pipe.branch_flush);
#endif

    pipe.PC = pipe.branch_dest;
  
    if(pipe.branch_flush >= 2) {
      if (pipe.decode_op) free(pipe.decode_op);
      pipe.decode_op = NULL; 
    }

    if (pipe.branch_flush >= 3) {
        if (pipe.execute_op) free(pipe.execute_op);
        pipe.execute_op = NULL;
    }

    if (pipe.branch_flush >= 4) {
        if (pipe.mem_op) free(pipe.mem_op);
        pipe.mem_op = NULL;
    }

    if (pipe.branch_flush >= 5) {
        if (pipe.wb_op) free(pipe.wb_op);
        pipe.wb_op = NULL;
    }

    pipe.branch_recover = 0;
    pipe.branch_dest = 0;
    pipe.branch_flush = 0;

    stat_squash++;
  }
}

/* Function to perform branch recovery */
void pipe_recover(int flush, uint32_t dest)
{
  /* if recovery already scheduled, override this recovery */
  if(pipe.branch_recover) return;
  
  pipe.branch_recover = 1;
  pipe.branch_flush = flush;
  pipe.branch_dest = dest;
}

void pipe_stage_wb()
{
  /* if no instruction in this pipeline stage, done */
  if(!pipe.wb_op) return;
  
  /* get op from pipeline */
  Pipe_Op *op = pipe.wb_op;
  pipe.wb_op = NULL;
  
  /* if the instruction writes to  register, do so now */
  /* Omits any use of PC as destination reg (except fro MOV) */
  if (op->reg_dst != -1 && op->reg_dst != 15) {
    pipe.REGS[op->reg_dst] = op->reg_dst_value;
#ifdef DEBUG
    printf("R%d = %08x\n", op->reg_dst, op->reg_dst_value);
#endif
  }
  /* if link, store lnker addr */
  if(op->link && op->is_branch) pipe.REGS[14] = op->pc+4;
  if(op->s) pipe.CPSR = op->CPSR;
  if(op->is_swi) {
    if(op->reg_src1 == 0xA) {
      pipe.PC = op->pc;
      RUN_BIT = 0;
    }  
  }
  free(op);

  stat_inst_retire++;
}

void pipe_stage_mem()
{
  /* if there is no inst in this pipeline stage, we're done */
  if (!pipe.mem_op) return;
  
  Pipe_Op *op = pipe.mem_op;
  
  if(op->is_mem)
  {
    switch(op->mem_flag & 0x5) {
      case LDRB: 
        op->reg_dst_value = mem_read_8(op->mem_addr);
        break;
      case LDR:
        if(op->reg_dst == 15) 
          op->reg_dst_value = mem_read_32(op->mem_addr) & 0xFFFFFFFC;
        else
          op->reg_dst_value = mem_read_32(op->mem_addr);
        break;
      case STRB:
        mem_write_8(op->mem_addr, op->reg_dst_value);
        op->reg_dst = -1;
        break;
      case STR:
	mem_write_32(op->mem_addr, op->reg_dst_value);
        op->reg_dst = -1;
        break;
    }
    op->data_ready = 1;
  }
  
  pipe.mem_op = NULL;
  pipe.wb_op = op;
}

void pipe_stage_ex()
{
  uint64_t temp_64_val = 0;
  uint32_t op2, temp_val;
  uint32_t C_flag;

  /* if a multiply is in progress, decrement cycles until value is ready */
  if (pipe.multiplier_stall > 0) pipe.multiplier_stall--;
  
  /* if downstream stall, return (and leave any input we had) */
  if (pipe.mem_op != NULL) return;
  
  /* if no execute op, return */
  if (pipe.execute_op == NULL) return;
  
  Pipe_Op *op = pipe.execute_op;

  if (!cond_is_met(op)) {
    pipe.execute_op= NULL;
    free(op);
    return;
  }
  switch (op->type)
  {
    case DATA_PROC:
      /* MUL/MLA */
      if (op->is_mul) {
        op->reg_dst_value = (op->reg_src1_value * op->reg_src2_value) & 0xFFFFFFFF; 
        if(op->accum) {
          op->reg_dst_value = (op->reg_dst_value + op->reg_src3_value) & 0xFFFFFFFF;
        }
        if(op->s) {
          set_N((op->reg_dst_value >> 31) & 0x1);
          set_Z(!op->reg_dst_value);
        }
        break;
      }
      else if (op->imm_bit) op2 = op->imm;
      else op2 = op->reg_src2_value; 
      switch(op->opcode) {
        case AND:
          op->reg_dst_value = op->reg_src1_value & op2;
          if(op->s) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(op->shifter_carry_out);
          }
          break;
        case EOR:
          op->reg_dst_value = (op->reg_src1_value)^op2;
	  if(op->s && op->reg_dst != 15) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!(op->reg_dst_value));
            set_C(op->shifter_carry_out);
          }
          break;
        case SUB:
          op->reg_dst_value = op->reg_src1_value - op2;
          if(op->s && (op->reg_dst != 15)) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(!(op2>op->reg_src1_value));
            if((((op->reg_src1_value >> 31) & 0x1) != ((op2 >> 31) & 0x1)) &&
               (((op->reg_src1_value >> 31) & 0x1) != ((op->reg_dst_value >> 31) & 0x1)))
              set_V(1);
            else set_V(0);
          }
          break;
        case RSB:
          op->reg_dst_value = op2 - op->reg_src1_value;
          if(op->s && (op->reg_dst != 15)) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(!(op->reg_src1_value>op2));
            if((((op->reg_src1_value >> 31) & 0x1) != ((op2 >> 31) & 0x1)) &&
               (((op->reg_src2_value >> 31) & 0x1) != ((op->reg_dst_value >> 31) & 0x1)))
              set_V(1);
            else set_V(0);
          }
          break;
        case ADD:
          op->reg_dst_value = op->reg_src1_value + op2;
          if(op->s && op->reg_dst != 15) {
            temp_64_val = (uint64_t) op->reg_src1_value + (uint64_t) op2;
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C((temp_64_val > (uint64_t) 0xFFFFFFFF));
            if((((op->reg_src1_value >> 31) & 0x1) == ((op2 >> 31) & 0x1)) &&
              (((op->reg_src1_value >> 31) & 0x1) != ((op->reg_dst_value>> 31) & 0x1)))
              set_V(1);
            else set_V(0);
          }
          break;
        case ADC:
          C_flag = get_C();
          op->reg_dst_value = op->reg_src1_value + op2 + C_flag;
          if(op->s && op->reg_dst != 15) {
            temp_64_val = (uint64_t) op->reg_src1_value + (uint64_t) op2 + C_flag;
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C((temp_64_val > (uint64_t) 0xFFFFFFFF));
            if((((op->reg_src1_value >> 31) & 0x1) == ((op2 >> 31) & 0x1)) &&
               (((op->reg_src1_value >> 31) & 0x1) != ((op->reg_dst_value>> 31) & 0x1)))
              set_V(1);
            else set_V(0);
          }
          break;
        case SBC:
          C_flag = get_C();
          op->reg_dst_value = op->reg_src1_value + ~(op2)+ C_flag;
          if(op->s && (op->reg_dst != 15)) {
            temp_64_val = (uint64_t) op->reg_src1_value + (uint64_t) ~(op2) + C_flag;
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C((temp_64_val > (uint64_t) 0xFFFFFFFF));
            if((((op->reg_src1_value >> 31) & 0x1) == ((~(op2) >> 31) & 0x1)) &&
              (((op->reg_src1_value >> 31) & 0x1) != ((op->reg_dst_value >> 31) & 0x1)))
                set_V(1);
                else set_V(0);
          }
          break;
        case RSC:
          C_flag = get_C();
          op->reg_dst_value = ~(op->reg_src1_value) + op2 + C_flag;
          if(op->s && (op->reg_dst != 15)) {
            temp_64_val = (uint64_t) ~(op->reg_src1_value) + (uint64_t) op2 + C_flag;
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C((temp_64_val > (uint64_t) 0xFFFFFFFF));
            if((((~(op->reg_src1_value) >> 31) & 0x1) == ((op2 >> 31) & 0x1)) &&
               (((op2 >> 31) & 0x1) != ((op->reg_dst_value >> 31) & 0x1)))
                set_V(1);
                else set_V(0);
            }
          break;
        case TST:
          temp_val = op->reg_src1_value & op2;
          set_N((temp_val >> 31) & 0x1);
          set_Z(!temp_val);
          set_C(op->shifter_carry_out);
          break;
        case TEQ:
          temp_val = op->reg_src1_value ^ op2;
          set_N((temp_val >> 31) & 0x1);
          set_Z(!temp_val);
          set_C(op->shifter_carry_out);
          break;
        case CMP:
          temp_val = op->reg_src1_value - op2;
          set_N((temp_val >> 31) & 0x1);
          set_Z(!temp_val);
          set_C(!(op2>op->reg_src1_value));
          if((((op->reg_src1_value >> 31) & 0x1) != ((op2 >> 31) & 0x1)) &&
            (((op->reg_src1_value >> 31) & 0x1) != ((temp_val>> 31) & 0x1)))
            set_V(1);
          else set_V(0);
          break;
        case CMN:
          temp_val = op->reg_src1_value+op2;
          temp_64_val = (uint64_t) op->reg_src1_value + (uint64_t) op2;
          set_N((temp_val >> 31) & 0x1);
          set_Z(!temp_val);
          set_C((temp_64_val > ((uint64_t) 0xFFFFFFFF)));
          if((((op->reg_src1_value >> 31) & 0x1) == ((op2 >> 31) & 0x1)) &&
             (((op->reg_src1_value >> 31) & 0x1) != ((temp_val >> 31) & 0x1)))
            set_V(1);
          else set_V(0);
          break;
        case ORR:
          op->reg_dst_value = op->reg_src1_value | op2;
          if(op->s && (op->reg_dst != 15)) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(op->shifter_carry_out);
          }
          break;
        case MOV:
          op->reg_dst_value = op2;
	  if(op->s && (op->reg_dst != 15)) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(op->shifter_carry_out);
          }
	  if(op->is_branch)
	    {
	      op->branch_taken=1;
	      op->branch_dest=op->target;
	      op->link = 0;
	    }
          break;
        case BIC:
          op->reg_dst_value = op->reg_src1_value & ~op2;
          if (op->s && (op->reg_dst_value != 15)) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(op->shifter_carry_out);
          }
          break;
        case MVN:
          op->reg_dst_value = ~op2;
          if(op->s && (op->reg_dst != 15)) {
            set_N((op->reg_dst_value >> 31) & 0x1);
            set_Z(!op->reg_dst_value);
            set_C(op->shifter_carry_out);
          }
          break;
        }
      op->data_ready = 1;
      break;
    case LD_STR:
      op->mem_addr = op->reg_src1_value;
      if(op->reg_src1 == 15) op->mem_addr += 8;
      if(!op->imm_bit && (op->mem_flag & 0x8))
        op->mem_addr += op->mem_imm;
      else if(!op->imm_bit)
        op->mem_addr -= op->mem_imm;
      else if(op->mem_flag & 0x8)
        op->mem_addr += op->reg_src2_value;
      else
        op->mem_addr -= op->reg_src2_value;
      break; 
    case BRANCH:
      op->branch_dest = op->pc + op->target;
      op->branch_taken = 1;
      break;
    default:
      break;        
  }
  
  if (op->branch_taken)
    pipe_recover(3, op->branch_dest);
  
  pipe.execute_op = NULL;
  pipe.mem_op = op;
}

void pipe_stage_decode()
{
  /* if downstream stall, return (and leave any input from before) */
  if (pipe.execute_op != NULL)
    return;

  /* if no op to decode, return */
  if (pipe.decode_op == NULL)
    return;

  /* grab op and remove from stage input */
  Pipe_Op *op = pipe.decode_op;
  //  pipe.decode_op = NULL;
  
  /* decoding an instruction */
  op->opcode = (op->instruction >> 21) & 0xF;
  uint32_t rs = (op->instruction >> 8) & 0xF;
  uint32_t rn = (op->instruction >> 16) & 0xF; 
  uint32_t rd = (op->instruction >> 12) & 0xF;
  uint32_t rm = op->instruction & 0xF;
  uint32_t imm = op->instruction & 0xFF;
  op->mem_imm = op->instruction & 0xFFF;
  uint32_t mul = (op->instruction >> 4) & 0xF;

  uint32_t shamt = (op->instruction >> 7) & 0x1F;
  op->cond = (op->instruction >> 28) & 0xF;
  op->type = (op->instruction >> 26) & 0x3;
  uint32_t shift_type = (op->instruction >> 5) & 0x3;
  uint32_t reg_shift = (op->instruction >> 4) & 0x1;

  op->imm_bit = (op->instruction >> 25) & 0x1; 
  op->mem_flag = (op->instruction >> 20) & 0x1F;
  op->accum = (op->instruction >> 21) & 0x1;
  uint32_t s = (op->instruction >> 20) & 0x1;
  uint32_t link = (op->instruction >> 24) & 0x1;

  uint32_t target = (op->instruction & 0xFFFFFF);
  uint32_t rot = rs;
  int stall = 0;

  if(pipe.mem_op && pipe.mem_op->s) op->CPSR = pipe.mem_op->CPSR;
  else if(pipe.wb_op && pipe.wb_op->s) op->CPSR = pipe.wb_op->CPSR; /* Perhaps unecessary */
  else op->CPSR = pipe.CPSR;

  op->reg_src1 = rn;

  /* if not multiply, check deps */
  if (!(op->type == DATA_PROC && !op->imm_bit && (mul == 9))) 
    {
      if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src1) 
	{
	  if(!pipe.mem_op->data_ready)
	    stall = 1;
	  else
	    op->reg_src1_value = pipe.mem_op->reg_dst_value;
	}
      else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src1) {
	op->reg_src1_value = pipe.wb_op->reg_dst_value;
      } 
      else
	op->reg_src1_value = pipe.REGS[rn];
    }
  
  if(stall) return;
  
  switch (op->type)
    {
    case DATA_PROC:
      /* MUL/MLA */
      if (!op->imm_bit && mul == 9) {
        op->reg_src1 = rs;
        op->reg_src2 = rm;
        op->reg_src3 = rd;
	op->s = s;
	op->is_mul = 1;

        /* check on op1 */
        if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src1) 
	  {
	    if(!pipe.mem_op->data_ready)
	      stall = 1;
	    else
	      op->reg_src1_value = pipe.mem_op->reg_dst_value;
	  }
        else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src1) 
	  {
	    op->reg_src1_value = pipe.wb_op->reg_dst_value;
	  } 
        else
          op->reg_src1_value = pipe.REGS[rs];
	
        /* check on op2 */
        if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src2) 
	  {
	    if(!pipe.mem_op->data_ready)
	      stall = 1;
	    else
	      op->reg_src2_value = pipe.mem_op->reg_dst_value;
	  }
        else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src2) 
	  {
	    op->reg_src2_value = pipe.wb_op->reg_dst_value;
	  } 
        else
          op->reg_src2_value = pipe.REGS[rm];
	
        /* check on op3 */
        if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src3) 
	  {
	    if(!pipe.mem_op->data_ready)
	      stall = 1;
	    else
	      op->reg_src3_value = pipe.mem_op->reg_dst_value;
	  }
        else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src3) 
	  {
	    op->reg_src3_value = pipe.wb_op->reg_dst_value;
	  } 
        else
          op->reg_src3_value = pipe.REGS[rd];
        
        if(stall) return;
        
        op->reg_dst = rn;
      }
      else {
      /*Calculate second operand, first parse type, then shift*/
	if (op->imm_bit) {
	  op->imm = ((imm >>(2*rot)) | (imm << (32-(rot*2))));
	  if (rot == 0)
	    op->shifter_carry_out= get_C(); /* STALL until CPSR is ready*/
	  else
	    op->shifter_carry_out = (op->imm >> 31) & 0x1;
	}
	else {
	  op->reg_src2 = rm;
	  
	  /* dep check */
          if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src2) {
            if(!pipe.mem_op->data_ready)
              stall = 1;
            else
              op->reg_src2_value = pipe.mem_op->reg_dst_value;
          }
          else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src2) {
            op->reg_src2_value = pipe.wb_op->reg_dst_value;
          } 
          else
            op->reg_src2_value = pipe.REGS[rm];
        
	  if(reg_shift) {
            /* dep check */
            if(pipe.mem_op && pipe.mem_op->reg_dst == rs) {
              if(!pipe.mem_op->data_ready)
                stall = 1;
              else
                shamt = pipe.mem_op->reg_dst_value;
              }
            else if(pipe.wb_op && pipe.wb_op->reg_dst == rs) {
              shamt = pipe.wb_op->reg_dst_value;
            } 
            else
              shamt = pipe.REGS[rs];
          }
 	        
	  if(stall) return;
 	  
	  switch(shift_type) {
	  case LSL:
	    if(shamt < 32) { 
	      op->shifter_carry_out = !shamt ? get_C() : ((op->reg_src2_value >>(32-shamt)) & 0x1);
	      op->reg_src2_value = op->reg_src2_value << shamt;
	    }
	    else if(shamt == 32) {
	      op->shifter_carry_out = op->reg_src2_value & 0x1;
	      op->reg_src2_value = 0;
	    }
	    else {
	      op->reg_src2_value = 0;
	      op->shifter_carry_out = 0;
	    }
	    break;
	    
	  case LSR:
	    if(!shamt && !reg_shift) {
	      op->shifter_carry_out = (op->reg_src2_value >> 31) & 0x1;
	      op->reg_src2_value = 0;
	    }
	    else if(!shamt) {
	      op->shifter_carry_out = get_C();
	    }
	    else if(shamt < 32) {
	      op->shifter_carry_out = ((op->reg_src2_value>>(shamt-1)) & 0x1);
	      op->reg_src2_value = op->reg_src2_value >> shamt;
	    }
	    else if(shamt == 32) {
	      op->shifter_carry_out =  (op->reg_src2_value >> 31) & 0x1;
	      op->reg_src2_value = 0;
	    }
	    else {
	      op->reg_src2_value = 0;
	      op->shifter_carry_out = 0;
	    }
	    break;
	  case ASR:
	    if (!shamt && !reg_shift) {
	      if (!((op->reg_src2_value >> 31) & 0x1)) {
		op->shifter_carry_out = (op->reg_src2_value >> 31) & 0x1;
		op->reg_src2_value = 0;
	      }
	      else {
		op->shifter_carry_out = (op->reg_src2_value >>  31) & 0x1;
		op->reg_src2_value = 0xFFFFFFFF;
	      }
	    }
	    else if(!shamt) {
	      op->shifter_carry_out = get_C();
	    }
	    else if(shamt < 32) {
	      op->shifter_carry_out = ((op->reg_src2_value>>(shamt-1)) & 0x1);
	      op->reg_src2_value = ((int)op->reg_src2_value) >> shamt;
	    }
	    else {
	      op->shifter_carry_out = (op->reg_src2_value >> 31) & 0x1;
	      if (!((op->reg_src2_value >> 31) & 0x1)) op->reg_src2_value = 0;
	      else op->reg_src2_value = 0xFFFFFFFF;
	    }
	    break;
	  case ROR:
	    if (!shamt && !reg_shift) {
	      op->shifter_carry_out = op->reg_src2_value & 0x1;
	      op->reg_src2_value = (get_C() << 31) | (op->reg_src2_value >> 1);
	    }
	    else if (!shamt) {
	      op->shifter_carry_out = get_C();
	    }
	    else if(!(shamt & 0x1F)) {
	      op->shifter_carry_out = (op->reg_src2_value >> 31) & 0x1;
	    }
	    else if (!reg_shift) {
	      op->shifter_carry_out = (op->reg_src2_value >> (shamt-1)) & 0x1;
	      op->reg_src2_value = ((op->reg_src2_value >> shamt) | (op->reg_src2_value << (32-shamt)));
	    }
	    else {
	      if(shamt == 32) {
		op->shifter_carry_out = (op->reg_src2_value >> 31) &0x1;
		shamt = shamt & 0x1F;
	      }
	      else {
		shamt = shamt & 0x1F;
		op->shifter_carry_out = (op->reg_src2_value >> (shamt-1)) & 0x1;
	      }
	      op->reg_src2_value = ((op->reg_src2_value >> shamt) | (op->reg_src2_value << (32-shamt)));
	    }
	    break;
	  }
	}
	switch(op->opcode) {
	case MOV:
	  if(rd == 15) {
	    op->is_branch = 1;
	    op->target = op->reg_src2_value;
	  }
	case AND:      
	case EOR:
	case SUB:
	case RSB:
      	case ADD:
      	case ADC:
        case SBC:
      	case RSC:
      	case ORR:
      	case BIC:
      	case MVN:
      	  op->reg_dst = rd;
	  op->s = s;
	  break;
        case TST:
        case TEQ:
        case CMP:
        case CMN:
	  op->s = 1;
	  break;
	}
      }
      break;
    case LD_STR:
      op->reg_src1 = rn;
      if(op->imm_bit) op->reg_src2 = rm;
      op->reg_dst = rd;
      op->is_mem = 1;

      if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src1) {
	if(!pipe.mem_op->data_ready)
	  stall = 1;
	else
	  op->reg_src1_value = pipe.mem_op->reg_dst_value;
      }
      else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src1) {
	op->reg_src1_value = pipe.wb_op->reg_dst_value;
      } 
      else
	op->reg_src1_value = pipe.REGS[rn];
      
      /* check on op2 */
      if(!op->imm_bit) 
	{
	  if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_src2) {
	    if(!pipe.mem_op->data_ready)
	      stall = 1;
	    else
	      op->reg_src2_value = pipe.mem_op->reg_dst_value;
	  }
	  else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_src2) {
	    op->reg_src2_value = pipe.wb_op->reg_dst_value;
	  } 
	  else
	    op->reg_src2_value = pipe.REGS[rm];
	}
      if ((op->mem_flag&0x5) == STRB || (op->mem_flag&0x5) == STR)
	{
 	  if(pipe.mem_op && pipe.mem_op->reg_dst == op->reg_dst) {
	    if(!pipe.mem_op->data_ready)
	      stall = 1;
	    else
	      op->reg_dst_value = pipe.mem_op->reg_dst_value;
	  }
	  else if(pipe.wb_op && pipe.wb_op->reg_dst == op->reg_dst) {
	    op->reg_dst_value = pipe.wb_op->reg_dst_value;
	  } 
	  else {
	    op->reg_dst_value = pipe.REGS[rd];
	  }
	}
      if(stall) return;
      break;
    case BRANCH:
      if(link) op->link = 1;
      op->is_branch = 1;
      op->target = (sign_extend_24to30((int) target) << 2)+8;
      break;
    case SWI:
      op->is_swi = 1;
      op->reg_src1 = rm;
      break;
    }
  
  pipe.decode_op = NULL;
  pipe.execute_op = op;
  return;
}

void pipe_stage_fetch()
{
  /* if pipeline is stalled, return*/
  if (pipe.decode_op != NULL)
    return;
  
  /* Allocate an op and send it down the pipeline. */
  Pipe_Op *op = malloc(sizeof(Pipe_Op));
  memset(op, 0, sizeof(Pipe_Op));
  op->reg_src1 = op->reg_src2 = op->reg_src3 = op->reg_dst = -1;

  op->instruction = mem_read_32(pipe.PC);
  op->pc = pipe.PC;
  pipe.decode_op = op;

  /* update PC*/
  pipe.PC += 4;
  stat_inst_fetch++;
}
