/*
 * ECE 18-447, Spring 2013
 *
 * MIPS pipeline timing simulator
 *
 * Chris Fallin, 2012
 */

#ifndef _PIPE_H_
#define _PIPE_H_

#include "shell.h"

/* Pipeline ops (instances of this structure) are high-level representations of
 * the instructions that actually flow through the pipeline. This struct does
 * not correspond 1-to-1 with the control signals that would actually pass
 * through the pieline. Rather, it carries the orginal instruction, operand
 * information and values as they are collected, and destination information. */
typedef struct Pipe_Op {
  /* PC of this instruction */
  uint32_t pc;
  /* raw instruction */
  uint32_t instruction;
  
  /* memory fields */
  uint32_t is_mem, mem_flag, mem_addr, mem_imm;
  
  /* branch flags */
  uint32_t is_branch, link, branch_dest, branch_taken, target;

  /* cond flag, CPSR */
  uint32_t s, CPSR, cond;

  /* other fields */
  uint32_t is_swi, is_mul, accum, type, shifter_carry_out;
  uint32_t imm_bit, imm, opcode;

  /* register source values */
  int reg_src1, reg_src2, reg_src3; /* 0 -- 15 if this inst has register source(s), or
			     -1 otherwise */
  uint32_t reg_src1_value, reg_src2_value, reg_src3_value; /* values of operands from source
					      regs */
  
  /* register destination information */
  int reg_dst; /* 0 -- 31 if this inst has a destination register, -1
		  otherwise */
  uint32_t reg_dst_value; /* value to write into dest reg. */
  int data_ready; /* destination value produced yet? */
  
} Pipe_Op;

/* The pipe state represents the current state of the pipeline. It holds a
 * pointer to the op that is currently at the input of each stage. As stages
 * execute, they remove the op from their input (set the pointer to NULL) and
 * place an op at their output. If the pointer that represents a stage's output
 * is not null when that stage executes, then this represents a pipeline stall,
 * and the stage must not overwrite its output (otherwise an instruction would
 * be lost).
 */

typedef struct Pipe_State {
  /* pipe op currently at the input of the given stage (NULL for none) */
  Pipe_Op *decode_op, *execute_op, *mem_op, *wb_op;
  
  /* register file state */
  uint32_t REGS[ARM_REGS];
  uint32_t CPSR;

  /* information for PC update (branch recovery). Branches should use this
   * mechanism to redirect the fetch stage, and flush the ops that came after
   * the branch as necessary. */
  int branch_recover; /* set to '1' to load a new PC */
  uint32_t branch_dest; /* next fetch will be from this PC */
  int branch_flush; /* how many stages to flush during recover? (1 = fetch, 2 = fetch/decode, ...) */
  
  /* multiplier stall info */
  int multiplier_stall; /* number of remaining cycles until HI/LO are ready */
  
  /* place other information here as necessary */
  
} Pipe_State;

/* global variable -- pipeline state */
extern Pipe_State pipe;

/* called during simulator startup */
void pipe_init();

/* this function calls the others */
void pipe_cycle();

/* helper: pipe stages can call this to schedule a branch recovery */
/* flushes 'flush' stages (1 = execute only, 2 = fetch/decode, ...) and then
 * sets the fetch PC to the given destination. */
void pipe_recover(int flush, uint32_t dest);

/* each of these functions implements one stage of the pipeline */
void pipe_stage_fetch();
void pipe_stage_decode();
void pipe_stage_ex();
void pipe_stage_mem();
void pipe_stage_wb();

#endif
